"""Copyright 2016-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
from __future__ import absolute_import


from builtins import str
from traceutils.record.record import Record


class InstructionRecord(Record):
    """Instruction trace record."""

    ADDR = "addr"
    ADDR_SIZE = "addr_size"
    COND_CODE = "cond_code"
    CPU = "cpu"
    CPU_EXEC_MODES = "svc|irq|fiq|usr|mon|sys|abt|und"
    DISASM = "disasm"
    INST_ID = "inst_id"
    INST_SET = "inst_set"
    MODE = "mode"
    OPCODE = "opcode"
    OPCODE_SIZE = "opcode_size"
    SECURITY = "security"

    INSN_FMT = r"(?P<%s>\d+) (?P<%s>\w+) (?P<%s>\w+) (?P<%s>(IT|IS))" \
               r" \((?P<%s>\d+)\) (?P<%s>[a-zA-Z0-9]+) (?P<%s>[a-zA-Z0-9]+)" \
               r" (?P<%s>A|T|X) (?P<%s>%s)_(?P<%s>s|ns) : (?P<%s>.+)"\
               % (Record.TIME, Record.SCALE, CPU, COND_CODE, INST_ID, ADDR,
                  OPCODE, INST_SET, MODE, CPU_EXEC_MODES, SECURITY, DISASM)

    def __init__(self, time, scale, cpu, cond_code_pass, inst_id, addr, opcode, inst_set, mode, sec,
                 disasm):
        """Default constructor."""
        Record.__init__(self, time, scale)
        self[self.CPU] = cpu
        self[self.COND_CODE] = cond_code_pass
        self[self.INST_ID] = int(inst_id)
        self[self.ADDR] = int(addr, 16)
        self[self.ADDR_SIZE] = len(addr)
        self[self.OPCODE] = int(opcode, 16)
        self[self.OPCODE_SIZE] = len(opcode)
        self[self.INST_SET] = inst_set
        self[self.MODE] = mode
        self[self.SECURITY] = sec
        self[self.DISASM] = disasm

    def get_cpu(self):
        """Returns the processor that gave the instruction."""
        return self[self.CPU]

    def get_cond_code_pass(self):
        """Returns whether the instruction passed the condition code."""
        return self[self.COND_CODE]

    def get_inst_id(self):
        """Returns the tick count of this processor."""
        return self[self.INST_ID]

    def get_addr(self):
        """Returns the source address in hex format (virtual address)."""
        return self[self.ADDR]

    def get_addr_size(self):
        """Returns size in characters for the hex value of the instruction source address."""
        return self[self.ADDR_SIZE]

    def get_opcode(self):
        """Returns the 16-bit/32-bit hex operation code of the instruction."""
        return self[self.OPCODE]

    def get_opcode_size(self):
        """Returns size in characters for the hex value of the operation code."""
        return self[self.OPCODE_SIZE]

    def get_inst_set(self):
        """Returns the instruction set identifier."""
        return self[self.INST_SET]

    def get_mode(self):
        """Returns the processor execution mode (svc, irq, fiq, usr, mon, sys, abt, und)."""
        return self[self.MODE]

    def get_sec(self):
        """Returns the processor security state (s or ns)."""
        return self[self.SECURITY]

    def get_disasm(self):
        """Returns the disassembly of the instruction."""
        return self[self.DISASM]

    @staticmethod
    def get_format():
        """Returns regular expression that encodes instruction record format."""
        return InstructionRecord.INSN_FMT

    def to_string(self):
        """Returns the string representation of this record, None fields are removed."""
        field_list = [self.get_time(), self.get_scale(), self.get_cpu(), self.get_cond_code_pass(),
                      self.get_inst_id(),
                      ("%x" % self.get_addr()).zfill(self.get_addr_size()),
                      ("%x" % self.get_opcode()).zfill(self.get_opcode_size()), self.get_inst_set(),
                      "_".join([self.get_mode(), self.get_sec()]), ":", self.get_disasm()]
        return " ".join(str(item) for item in field_list)
