"""Copyright 2016-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
from __future__ import absolute_import


from builtins import str
from traceutils.record.record import Record


class MemoryAccessRecord(Record):
    """Memory access trace record."""

    ADDR = "addr"
    ADDR_SIZE = "add_size"
    ATTRIB = "attrib"
    DATA = "data"
    DATA_SIZE = "data_size"
    RW = "rw"
    SIZE = "size"

    MEMORY_ACCESS_FMT = r"(?P<%s>\d+) (?P<%s>\w+) M(?P<%s>R|W)(?P<%s>1|2|4|8)" \
                        r"(?P<%s>(X|T|L))? (?P<%s>[a-zA-Z0-9]+) (?P<%s>[a-zA-Z0-9]+)" \
                        % (Record.TIME, Record.SCALE, RW, SIZE, ATTRIB, ADDR, DATA)

    def __init__(self, time, scale, rw, size, attrib, addr, data):
        """Default constructor."""
        Record.__init__(self, time, scale)
        self[self.RW] = rw
        self[self.SIZE] = int(size)
        self[self.ATTRIB] = attrib
        self[self.ADDR] = int(addr, 16)
        self[self.ADDR_SIZE] = len(addr)
        self[self.DATA] = int(data, 16)
        self[self.DATA_SIZE] = len(data)

    def get_rw(self):
        """Returns read/write access type."""
        return self[self.RW]

    def get_size(self):
        """Returns the size of the data transfer in bytes (1, 2, 4, 8)."""
        return self[self.SIZE]

    def get_attrib(self):
        """Returns the optional access attribute."""
        return self[self.ATTRIB]

    def get_addr(self):
        """Returns the virtual address used to access memory in hexadecimal format."""
        return self[self.ADDR]

    def get_addr_size(self):
        """Returns size in characters for the hexadecimal value of the virtual address."""
        return self[self.ADDR_SIZE]

    def get_data(self):
        """Returns the hexadecimal value of the transferred data."""
        return self[self.DATA]

    def get_data_size(self):
        """Returns size in characters for the hexadecimal value of the transferred data."""
        return self[self.DATA_SIZE]

    @staticmethod
    def get_format():
        """Returns regular expression that encodes memory access record format."""
        return MemoryAccessRecord.MEMORY_ACCESS_FMT

    def to_string(self):
        """Returns the string representation of this record, None fields are removed."""
        field_list = [self.get_time(), self.get_scale(),
                      "".join(["M", self.get_rw(), str(self.get_size()),
                               "" if self.get_attrib() is None else self.get_attrib()]),
                      ("%x" % self.get_addr()).zfill(self.get_addr_size()),
                      ("%x" % self.get_data()).zfill(self.get_data_size())]
        return " ".join(str(item) for item in field_list)
